/* $Id: motd.c,v 1.1.1.1 1995/07/22 00:57:18 sverrehu Exp $ */
/**************************************************************************
 *
 *  FILE            motd.c
 *
 *  MODULE OF       motd
 *
 *  DESCRIPTION     
 *
 *
 *  WRITTEN BY      Sverre H. Huseby
 *
 *  MODIFICATIONS   
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <locale.h>

#include "misc.h"
#include "motd.h"

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

typedef struct {
    time_t t;
    char  *msg;
} message;

static message **Msg = NULL;
static int nMsg = 0;
static char *theMessage = NULL;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static message *newMessage(time_t t, char *msg)
{
    message *ret;
    
    ret = xmalloc(sizeof(message));
    ret->t = t;
    ret->msg = xmalloc(strlen(msg) + 1);
    strcpy(ret->msg, msg);
    return ret;
}

static void strext(char **dest, char *src)
{
    if (*dest == NULL) {
	*dest = xmalloc(strlen(src) + 1);
	strcpy(*dest, src);
	return;
    }
    *dest = xrealloc(*dest, strlen(*dest) + strlen(src) + 1);
    strcat(*dest, src);
}

static void addMessage(message *m)
{
    if (Msg == NULL) {
	Msg = xmalloc(sizeof(message *));
	Msg[0] = m;
	nMsg = 1;
	return;
    }
    Msg = xrealloc(Msg, (nMsg + 1) * sizeof(message *));
    Msg[nMsg++] = m;
}

static time_t scanDate(char *s, struct tm *st)
{
    struct tm t;
    int t1, t2, t3;

    t1 = t2 = t3 = 0;
    while (isspace(*s))
	++s;
    if (!isdigit(*s))
        return 0;
    while (isdigit(*s))
        t1 = 10 * t1 + *s++ - '0';
    if (*s != '/' && *s != '-')
        return 0;
    
    if (!isdigit(*++s))
        return 0;
    while (isdigit(*s))
        t2 = 10 * t2 + *s++ - '0';
    if (*s != '/' && *s != '-')
        return 0;

    if (!isdigit(*++s))
        return 0;
    while (isdigit(*s))
        t3 = 10 * t3 + *s++ - '0';

    t.tm_mday = t1;
    t.tm_mon = t2 - 1;
    t.tm_year = t3;

    if (t.tm_year >= 1900L)
        t.tm_year -= 1900L;
    else if (t.tm_year >= 0L && t.tm_year <= 37L)
        t.tm_year += 100L;

    while (isspace(*s))
	++s;
    if (*s == ',')
	++s;
    while (isspace(*s))
	++s;

    if (*s) {
	t1 = t2 = t3 = 0;
	if (!isdigit(*s))
	    return 0;

	while (isdigit(*s))
	    t1 = 10 * t1 + *s++ - '0';
	if (*s != ':' && *s != '.')
	    return 0;

	if (!isdigit(*++s))
	    return 0;
	while (isdigit(*s))
	    t2 = 10 * t2 + *s++ - '0';

	t3 = 0;
	if (*s) {
	    if (*s != ':' && *s != '.') {
		while (isspace(*s))
		    ++s;
		if (*s)
		    return 0;
	    } else {
		if (!isdigit(*++s))
		    return 0;
		while (isdigit(*s))
		    t3 = 10 * t3 + *s++ - '0';
		
		while (isspace(*s))
		    ++s;
		if (*s)
		    return 0;
	    }
	}
	t.tm_hour = t1;
	t.tm_min = t2;
	t.tm_sec = t3;
    } else
	t.tm_hour = t.tm_min = t.tm_sec = 0;

    if (st)
	*st = t;
    return mktime(&t);
}

static void readMessages(void)
{
    int lno = 0;
    FILE *f;
    char line[256], *l;
    message *msg = NULL;
    time_t t;
    struct tm st;

    if ((f = fopen(MESSAGE_FILE, "r")) == NULL)
	if (errno == ENOENT)
	    exit(0);
	else
	    FatalPerror(MESSAGE_FILE);
    while (fgets(line, 256, f) != NULL) {
	++lno;
	l = line;
	while (isspace(*l))
	    ++l;
	if (*l == COMMENT_CHAR)
	    continue;
	if (strncmp(line, DATE_PREFIX, strlen(DATE_PREFIX)) == 0) {
	    t = scanDate(line + 3, &st);
	    if (!t)
		Fatal("error in date in line %d of " MESSAGE_FILE, lno);
	    if (st.tm_hour || st.tm_min || st.tm_sec)
		strftime(line, 255, "%x %X\n", &st);
	    else
		strftime(line, 255, "%x\n", &st);
	    msg = newMessage(t, line);
	    addMessage(msg);
	} else {
	    strext(msg ? &(msg->msg) : &theMessage, line);
	}
    }
    fclose(f);
}

static int messageCmp(const message **m1, const message **m2)
{
    if ((*m1)->t > (*m2)->t)
	return -1;
    else if ((*m1)->t < (*m2)->t)
	return 1;
    return 0;
}

static void sortMessages(void)
{
    if (!Msg)
	return;
    qsort(Msg, nMsg, sizeof(message *),
	  (int (*)(const void *, const void *)) messageCmp);
}

static char *getDateFile(void)
{
    char *h;
    static char fname[256];

    if ((h = getenv("HOME")) == NULL)
	Fatal("no $HOME-variable");
    strcpy(fname, h);
    strcat(fname, "/");
    strcat(fname, DATE_FILE);
    return fname;
}

static time_t getDateLastRead(void)
{
    struct stat st;

    if (stat(getDateFile(), &st) == -1)
	return 0;
    return st.st_ctime;
}

static void touchDateFile(void)
{
    int fd;

    if ((fd = creat(getDateFile(), 0666)) != -1)
	close(fd);
}

static int addMessages(time_t last)
{
    int q, ret = 0;

    for (q = 0; q < nMsg; q++)
	if (Msg[q]->t > last) {
	    strext(&theMessage, Msg[q]->msg);
	    ++ret;
	} else
	    break;
    return ret;
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void main(int argc, char *argv[])
{
    time_t last;

    /* Maybe this fork() will avoid zombies... */
    if (fork())
	return;
    
    SetProgName(argv[0]);
    setlocale(LC_ALL, "");
    last = getDateLastRead();
    readMessages();
    sortMessages();
    touchDateFile();
    if (addMessages(last))
	openWindow(argc, argv, theMessage);
}
